#include <stdio.h>
#include <stdlib.h>

/* #define TRUNC_A */

int
main (int argc, char *argv[])
{
  long i;
  unsigned long NTL_BITS_PER_LONG, K, CHOP, REM, MASK, fn, mask2;

  if (argc != 4)
    {
      fprintf (stderr, "Usage: gen_bb_mul_code NTL_BITS_PER_LONG K fn\n");
      fprintf (stderr, "       where fn is 0 for mul1, 1 for Mul1, and 2 for AddMul1.\n");
      exit (1);
    }
  
  NTL_BITS_PER_LONG = atoi (argv[1]);
  K = atoi (argv[2]);
  fn = atoi (argv[3]);

  CHOP = NTL_BITS_PER_LONG % K; /* first block */
  if (CHOP == 0)
    CHOP = K;
  REM = NTL_BITS_PER_LONG - K;
  MASK = (1UL << K) - 1UL;

  printf ("/* This file was generated automatically with\n");
  printf ("   gen_bb_mul_code %u %u %u. Don't edit it */\n\n",
          NTL_BITS_PER_LONG, K, fn);

  if (fn == 0)
    {
      //      fprintf (stderr, "Generating mul1() routine\n");
      printf ("static inline\n");
      printf ("void mul1(_ntl_ulong *c, _ntl_ulong a, _ntl_ulong b)\n");
    }
  else if (fn == 1)
    {
      //      fprintf (stderr, "Generating Mul1() routine\n");
      printf ("static _ntl_ulong\n");
      printf ("Mul1 (_ntl_ulong *cp, const _ntl_ulong *bp, long sb, _ntl_ulong a)\n");
    }
  else if (fn == 2)
    {
      // fprintf (stderr, "Generating AddMul1() routine\n");
      printf ("static _ntl_ulong\n");
      printf ("AddMul1 (_ntl_ulong *dp, const _ntl_ulong *cp, const _ntl_ulong* bp, long sb,\n");
      printf ("        _ntl_ulong a)\n");
    }
  else
    {
      fprintf (stderr, "Error: unexpected fn\n");
      exit (1);
    }
  printf ("{\n");
  if (fn > 0) /* Mul1/AddMul1 */
    {
      printf ("   long i;\n");
      printf ("   _ntl_ulong carry = 0, b;\n");
    }
  printf ("   _ntl_ulong hi, lo;\n");
  printf ("   _ntl_ulong A[%u];\n", 1<<K);
  printf ("\n");
  printf ("   A[0] = 0;\n");
#ifdef TRUNC_A /* truncate a so that all multiples will fit in a word */
  printf ("   A[1] = a & ((1UL << (NTL_BITS_PER_LONG-%u))-1UL);\n", K-1);
#else /* do not truncate: fix non-considered bit products afterwards */
  printf ("   A[1] = a;\n");
#endif
  for (i = 2; i < (1 << K); i++)
    {
      if (i % 2 == 0)
        printf ("   A[%u] = A[%u] << 1;\n", i, i / 2);
      else
        printf ("   A[%u] = A[%u] ^ A[1];\n", i, i - 1);
    }
  printf ("\n");

  if (fn > 0)
    {
      printf ("   for (i = 0; i < sb; i++)\n");
      printf ("     {\n");
      printf ("       b = bp[i];\n");
    }
  printf ("   lo = A[b >> (NTL_BITS_PER_LONG - %u)];\n", CHOP);
  printf ("   hi = lo >> (NTL_BITS_PER_LONG - %u);\n", K);
  printf ("   lo = (lo << %u) ^ A[(b >> (NTL_BITS_PER_LONG - %u)) & %u];\n",
          K, CHOP + K, MASK);
  for (i = NTL_BITS_PER_LONG - CHOP - 2 * K; i >= K; i -= K)
    {
      printf ("   hi = (hi << %u) | (lo >> %u);\n", K, REM);
      printf ("   lo = (lo << %u) ^ A[(b >> %u) & %u];\n", K, i, MASK);
    }
  /* special case for i=0 since a shift of 0 is undefined */
  printf ("   hi = (hi << %u) | (lo >> %u);\n", K, REM);
  printf ("   lo = (lo << %u) ^ A[b & %u];\n", K, MASK);
#ifndef TRUNC_A
  MASK = 0;
  for (i = 0; i < NTL_BITS_PER_LONG; i += K)
    MASK |= 1UL << i;
  MASK = ~MASK;
  mask2 = MASK;
#endif  
  for (i = 1; i < K; i++)
    {
      printf ("   if ((a >> (NTL_BITS_PER_LONG - %u)) & 1)\n", i);
#ifdef TRUNC_A
      printf ("     {\n");
      printf ("       hi = hi ^ (b >> %u);\n", i);
      printf ("       lo = lo ^ (b << (NTL_BITS_PER_LONG - %u));\n", i);
      printf ("     }\n");
#else
      /* bit NTL_BITS_PER_LONG-i from a was not considered in blocks of
         K bits from b for index j >= i */
      printf ("       hi = hi ^ ((b & 0x%lx) >> %u);\n", mask2, i);
      mask2 = (mask2 << 1) & MASK;
#endif
    }
  if (fn == 0)
    {
      printf ("   c[0] = lo;\n");
      printf ("   c[1] = hi;\n");
    }
  else if (fn == 1)
    {
      printf ("   cp[i] = carry ^ lo;\n");
      printf ("   carry = hi;\n");
    }
  else if (fn == 2)
    {
      printf ("   dp[i] = cp[i] ^ (carry ^ lo);\n");
      printf ("   carry = hi;\n");
    }
  if (fn > 0)
    {
      printf ("    }\n");
      printf ("   return carry;\n");
    }
  printf ("}\n");
}
